# -*- coding: utf-8 -*-
import argparse


class AppendReplace(argparse.Action):
    """argparse.Action used for appending values on the command-line"""

    def __call__(self, parser, namespace, values, option_string=None):
        # if we find an attribute and it is equal to the default,
        # then we need to completely replace the value
        current_value = getattr(namespace, self.dest, [])
        if current_value == self.default:
            setattr(namespace, self.dest, [values])
        else:
            current_value.append(values)
            setattr(namespace, self.dest, current_value)


class AppendBoolean(argparse.Action):
    """Action used for appending boolean values on the command-line"""

    def __init__(self,
                 option_strings,
                 dest,
                 const,
                 default=None,
                 required=False,
                 help=None,
                 metavar=None):
        super(AppendBoolean, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            const=const,
            default=default,
            required=required,
            help=help,
            metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        # if we find an attribute and it is equal to the default,
        # then we need to completely replace the value
        current_value = getattr(namespace, self.dest, [])
        if current_value == self.default:
            setattr(namespace, self.dest, [self.const])
        else:
            current_value.append(self.const)
            setattr(namespace, self.dest, current_value)


class MergeAction(argparse.Action):
    """Merges command-line values into a single dictionary based on separator.

    Each MergeAction has a child_action that indicates what should happen
    for each value. It uses the separator to determine the eventual
    location for each of its values.

    The dest is split up by separator and each string is in turn used to
    determine the key that should be used to store this value in the
    dictionary that will get created.

    Attributes:
        child_action: The action that determines which value is stored
        child_const: For booleans, this is the value used
        separator: A separator to split up keys in the dictionary
    """

    def __init__(self,
                 option_strings,
                 dest,
                 nargs=None,
                 const=None,
                 default=None,
                 type=None,
                 choices=None,
                 required=False,
                 help=None,
                 metavar=None,
                 child_action=None,
                 separator='.',
                 child_const=None):
        super(MergeAction, self).__init__(option_strings=option_strings,
                                          dest=dest,
                                          nargs=nargs,
                                          const=const,
                                          default=default,
                                          type=type,
                                          choices=choices,
                                          required=required,
                                          help=help,
                                          metavar=metavar,
                                          )
        self.child_action = child_action
        self.child_const = child_const
        self.separator = separator
        self.parent_names = self.dest.split(self.separator)

    def _get_leaf_dict_and_key(self, namespace):
        # Traverse the namespace down to the lowest dict and return
        # a dictionary and key that represents the "leaf" node
        if not hasattr(namespace, self.parent_names[0]):
            setattr(namespace, self.parent_names[0], {})

        current_dict = getattr(namespace, self.parent_names[0])

        for index, parent_name in enumerate(self.parent_names[1:]):
            if index == len(self.parent_names) - 2:
                return current_dict, parent_name
            elif parent_name not in current_dict:
                current_dict[parent_name] = {}

            current_dict = current_dict[parent_name]

        return current_dict, self.parent_names[-1]

    def _merge_value(self, leaf_dict, name, value):
        if self.child_action == 'store_true':
            leaf_dict[name] = True

        elif self.child_action == 'store_false':
            leaf_dict[name] = False

        elif self.child_action == AppendBoolean:
            if name in leaf_dict:
                leaf_dict[name].append(self.child_const)
            else:
                leaf_dict[name] = [self.child_const]

        elif self.child_action == 'store':
            leaf_dict[name] = value

        elif self.child_action == AppendReplace:
            if name in leaf_dict:
                leaf_dict[name].append(value)
            else:
                leaf_dict[name] = [value]

        else:
            raise ValueError("Don't know how to do action: {0}"
                             .format(self.child_action))

    def __call__(self, parser, namespace, values, option_string=None):
        if hasattr(namespace, self.dest):
            delattr(namespace, self.dest)
        leaf_dict, name = self._get_leaf_dict_and_key(namespace)
        self._merge_value(leaf_dict, name, values)



# -*- coding: utf-8 -*-
from watchdog.events import RegexMatchingEventHandler

import yapconf
from yapconf.exceptions import YapconfLoadError, YapconfSourceError


class ConfigChangeHandler(object):
    """Handles config changes.

    Expects a watcher to call it when a particular config changes.
    """

    def __init__(self, current_config, spec, user_handler=None):
        self.current_config = current_config
        self.spec = spec
        self.user_handler = user_handler

    def handle_config_change(self, new_config):
        """Handle the new configuration.

        Args:
            new_config (dict): The new configuration

        """
        if self.user_handler:
            self.user_handler(self.current_config, new_config)
        self._call_spec_handlers(new_config)
        self.current_config = new_config

    def _call_spec_handlers(self, new_config):
        flattened_config = yapconf.flatten(new_config, self.spec._separator)
        flattened_current = yapconf.flatten(
            self.current_config, self.spec._separator
        )

        for key, value in flattened_config.items():
            if value != flattened_current.get(key):
                item = self.spec.find_item(key)
                if item and item.watch_target:
                    item.watch_target(flattened_current[key], value)


class FileHandler(RegexMatchingEventHandler):
    """Watchdog handler that only watches a specific file."""

    def __init__(self, filename, handler, file_type='json'):
        super(FileHandler, self).__init__(regexes=[r'^%s$' % filename])
        self._handler = handler
        self._filename = filename
        self._file_type = file_type

    def on_deleted(self, event):
        raise YapconfSourceError(
            'While watching file %s the file was deleted. Aborting watch.'
            % self._filename
        )

    def on_modified(self, event):
        new_config = self._load_config(self._filename)
        self._handler.handle_config_change(new_config)

    def _load_config(self, filename):
        return yapconf.load_file(
            filename,
            file_type=self._file_type,
            klazz=YapconfLoadError
        )



# -*- coding: utf-8 -*-
import pytest
from mock import Mock, patch

from yapconf.exceptions import YapconfSourceError
from yapconf.handlers import ConfigChangeHandler, FileHandler


def test_handle_config_change(simple_spec):
    current_config = {'my_string': 'foo'}
    new_config = {'my_string': 'bar'}
    user_handler = Mock()
    item_handler = Mock()

    handler = ConfigChangeHandler(
        current_config, simple_spec, user_handler
    )

    item = simple_spec.find_item('my_string')
    item.watch_target = item_handler

    handler.handle_config_change(current_config)
    user_handler.assert_called_with(current_config, current_config)
    assert item_handler.call_count == 0

    handler.handle_config_change(new_config)
    assert item_handler.call_count == 1


def test_handle_config_change_too_many_items(simple_spec):
    current_config = {'my_string': 'foo'}
    new_config = {'NOT_IN_SPEC': 'bar'}
    user_handler = Mock()
    item_handler = Mock()

    handler = ConfigChangeHandler(
        current_config, simple_spec, user_handler
    )

    item = simple_spec.find_item('my_string')
    item.watch_target = item_handler

    handler.handle_config_change(new_config)
    user_handler.assert_called_with(current_config, new_config)
    assert item_handler.call_count == 0


def test_file_handler():
    custom_handler = Mock()
    handler = FileHandler('filename', custom_handler)

    with patch('yapconf.load_file') as mock_load:
        mock_load.return_value = "new_config"
        handler.on_modified(None)
        custom_handler.handle_config_change.assert_called_with('new_config')


def test_file_handler_on_deleted():
    handler = FileHandler('filename', Mock())
    with pytest.raises(YapconfSourceError):
        handler.on_deleted(None)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from yapconf.items import YapconfDictItem

# flake8: noqa
HEADER = """# {app_name} Configuration

This document describes the configuration for {app_name}. Each section will 
document a particular configuration value and its description. First, 
though, we start with the possible sources. 

This documentation was auto-generated by [yapconf.](https://github.com/loganasherjones/yapconf)
"""

CONFIG_HEADER = """## Configuration

This section outlines the various configuration items {app_name} supports.
"""

SOURCES_HEADER = """
## Sources

{app_name} configuration can be loaded from the below sources:
"""

SOURCE_TEMPLATE = """
#### {source_label}

{source_label} is of type {source_type_link}.

{source_type_description}
"""

ITEM_TEMPLATE = """### {name}

{description}

{attribute_table}
{item_options}

{long_description}
"""

SOURCE_TYPE_LINKS = {
    'dict': 'https://yapconf.readthedocs.io/en/stable/sources.html#dict',
    'environment': 'https://yapconf.readthedocs.io/en/stable/sources.html#environment',
    'etcd': 'https://yapconf.readthedocs.io/en/stable/sources.html#etcd',
    'json': 'https://yapconf.readthedocs.io/en/stable/sources.html#json',
    'kubernetes': 'https://yapconf.readthedocs.io/en/stable/sources.html#kubernetes',
    'yaml': 'https://yapconf.readthedocs.io/en/stable/sources.html#yaml',
}

VALID_JSON_LINK = "[valid JSON](https://www.json.org/)"
VALID_YAML_LINK = "[valid YAML](http://yaml.org/)"

DICT_DESCRIPTION = """
The dict source means that {app_name} will load configuration values from an 
in-memory dictionary that cannot be modified externally. Usually, it is used as 
a list of sensible defaults."""

ENVIRONMENT_DESCRIPTION = """
This environment source means that {app_name} will load configuration values 
by using their `env_name` that is documented below."""

ETCD_DESCRIPTION = """
The etcd source means that {app_name} will load configuration values from the 
`{key}` in etcd."""

JSON_DATA_DESCRIPTION = """
This json source means that {app_name} will load configuration values from an 
in-memory string and cannot be modified. Usually it is used as a list of 
sensible defaults."""

JSON_FILE_DESCRIPTION = """
This json source means that {app_name} will load configuration values from 
`{filename}`. You may edit this file with {valid_link}."""

K8S_DESCRIPTION = """The kubernetes source indicates that the configuration 
may be loaded from a kubernetes [ConfigMap](https://kubernetes.io/docs/tasks/configure-pod-container/configure-pod-configmap/). 
Specifically, {app_name} will load the configuration from the ConfigMap named 
`{config_map_name}` in the namespace `{config_map_namespace}`"""

K8S_KEY_DESCRIPTION = """
Additionally, {source_label} specified that `{key}` exists in the 
`{config_map_name}` ConfigMap. `{key}` will be loaded as `{config_type}`. 
You may modify this entry with {valid_link}."""

YAML_DESCRIPTION = """
The yaml source means that {app_name} will load configuration values from 
`{filename}`. You may edit this file with {valid_link}"""


def _find_row_maxes(headers, rows):
    maxes = {}
    for key, value in headers.items():
        maxes[key] = len(value)
    for row in rows:
        for key, value in row.items():
            maxes[key] = max(maxes[key], len(value))
    return maxes


def _get_padding(key, row_maxes, value, pad_value=" ", modifier=0):
    return pad_value * (row_maxes[key] - len(value) + modifier)


def _build_row(row, row_maxes, row_keys, pad_value=" "):
    row_value = "|"
    for key in row_keys:
        padding = _get_padding(key, row_maxes, row[key], pad_value)
        row_value = row_value + ' ' + row[key] + padding + ' |'
    return row_value


def _build_separator(row_maxes, row_keys):
    fake_row = row_maxes.copy()
    for key, value in row_maxes.items():
        fake_row[key] = ''
    return _build_row(fake_row, row_maxes, row_keys, pad_value='-')


def build_markdown_table(headers, rows, row_keys=None):
    """Build a lined up markdown table.

    Args:
        headers (dict): A key -> value pairing fo the headers.
        rows (list): List of dictionaries that contain all the keys listed in
        the headers.
        row_keys (list): A sorted list of keys to display

    Returns:
        A valid Markdown Table as a string.
    """
    row_maxes = _find_row_maxes(headers, rows)
    row_keys = row_keys or [key for key, value in headers.items()]
    table = [
        _build_row(headers, row_maxes, row_keys),
        _build_separator(row_maxes, row_keys)
    ]

    for row in rows:
        table.append(_build_row(row, row_maxes, row_keys))
    return '\n'.join(table) + '\n'


def _generate_source_description(source, app_name, source_label):
    format_kwargs = {'app_name': app_name}
    template = ""
    if source.type == 'dict':
        template = DICT_DESCRIPTION

    elif source.type == 'environment':
        template = ENVIRONMENT_DESCRIPTION

    elif source.type == 'etcd':
        template = ETCD_DESCRIPTION
        format_kwargs['key'] = source.key

    elif source.type == 'json':
        if source.filename:
            template = JSON_FILE_DESCRIPTION
            format_kwargs['filename'] = source.filename
            format_kwargs['valid_link'] = VALID_JSON_LINK
        else:
            template = JSON_DATA_DESCRIPTION

    elif source.type == 'kubernetes':
        template = K8S_DESCRIPTION
        format_kwargs['config_map_name'] = source.name
        format_kwargs['config_map_namespace'] = source.namespace

        if source.key:
            template = "\n".join([t for t in [template, K8S_KEY_DESCRIPTION]])
            format_kwargs['source_label'] = source_label
            format_kwargs['key'] = source.key
            format_kwargs['config_type'] = source.config_type
            if source.config_type == 'yaml':
                format_kwargs['valid_link'] = VALID_YAML_LINK
            else:
                format_kwargs['valid_link'] = VALID_JSON_LINK

    elif source.type == 'yaml':
        template = YAML_DESCRIPTION
        format_kwargs['filename'] = source.filename
        format_kwargs['valid_link'] = VALID_YAML_LINK

    return template.format(**format_kwargs)


def _generate_source_section(source_label, source, app_name):
    source_type_link = (
        "[%s](%s)" % (source.type, SOURCE_TYPE_LINKS[source.type])
    )

    source_type_description = _generate_source_description(
        source, app_name, source_label
    )
    return SOURCE_TEMPLATE.format(
        source_label=source_label,
        source_type_link=source_type_link,
        source_type_description=source_type_description,
    )


def _generate_item_table(item):
    headers = {
        'attribute': 'Attribute',
        'value': 'Value',
    }
    if item.cli_support:
        cli_names = item.cli_names
    else:
        cli_names = None

    rows = [
        {
            'attribute': '**item_type**',
            'value': '`%s`' % item.item_type,
        },
        {
            'attribute': '**default**',
            'value': '`%s`' % item.default,
        },
        {
            'attribute': '**env_name**',
            'value': '`%s`' % item.env_name,
        },
        {
            'attribute': '**required**',
            'value': '`%s`' % item.required,
        },
        {
            'attribute': '**cli_name**',
            'value': '`%s`' % cli_names,
        },
        {
            'attribute': '**fallback**',
            'value': '`%s`' % item.fallback
        },
        {
            'attribute': '**choices**',
            'value': '`%s`' % item.choices,
        },
    ]
    return build_markdown_table(headers, rows, ['attribute', 'value'])


def _generate_item_options(item, app_name):
    options = []
    if item.env_name:
        options.append(
            'You can set {name} from the environment by setting the '
            'environment variable `{env_name}`'.format(
                name=item.fq_name,
                env_name=item.env_name
            )
        )

    if item.cli_support:
        cli_names = item.cli_names
        options.append(
            'You can set `{name}` from the command-line by specifying '
            '`{cli_names}` at {app_name}\'s entrypoint.'.format(
                name=item.fq_name,
                cli_names=cli_names,
                app_name=app_name,
            )
        )

    if item.fallback:
        option = (
            'If `{name}` is not set in any of the sources listed, it will '
            'attempt to fallback to the value set in `{fallback}`.'
        ).format(
            name=item.fq_name,
            fallback=item.fallback
        )

        if item.default:
            option += (
                ' If neither the fallback nor the original key is found '
                'then the value will fallback to the default of `{default}`'
            ).format(default=('%s' % item.default))

        options.append(option)
    elif item.default:
        options.append((
            'If `{name}` is not set in any of the sources listed, it will '
            'fallback to the default value `{value}`'.format(
                name=item.fq_name,
                value=item.default
            )
        ))

    return '\n\n'.join(options)


def _generate_item_section(item, app_name):
    item_table = _generate_item_table(item)
    item_options = _generate_item_options(item, app_name)
    long_description = item.long_description or ""
    return ITEM_TEMPLATE.format(
        name=item.fq_name,
        description=item.description or "No description provided.",
        attribute_table=item_table,
        item_options=item_options,
        long_description=long_description,
    )


def _get_table_row_from_item(item):
    return {
        'name': '[' + item.fq_name + '](#' + item.fq_name + ')',
        'type': item.item_type,
        'default': '%s' % item.default,
        'description': '%s' % item.description,
    }


def _generate_item_sections(items, app_name):
    rows = []
    item_sections = []
    for item in items:
        if isinstance(item, YapconfDictItem):
            tmp_rows, tmp_sections = _generate_item_sections(
                _sorted_dict_values(item.children),
                app_name
            )
            rows += tmp_rows
            item_sections += tmp_sections
        else:
            rows.append(_get_table_row_from_item(item))
            item_sections.append(_generate_item_section(item, app_name))

    return rows, item_sections


def _sorted_dict_values(dictionary):
    sorted_keys = sorted(list(dictionary))
    return [dictionary[key] for key in sorted_keys]


def generate_markdown_doc(app_name, spec):
    """Generate Markdown Documentation for the given spec/app name.

    Args:
        app_name (str): The name of the application.
        spec (YapconfSpec): A yapconf specification with sources loaded.

    Returns (str):
        A valid, markdown string representation of the documentation for
        the given specification.
    """

    # Apply standard headers.
    sections = [
        HEADER.format(app_name=app_name),
        SOURCES_HEADER.format(app_name=app_name)
    ]

    # Generate the sources section of the documentation
    sorted_labels = sorted(list(spec.sources))
    for label in sorted_labels:
        sections.append(
            _generate_source_section(label, spec.sources[label], app_name)
        )

    # Generate the config section.
    sections.append(CONFIG_HEADER.format(app_name=app_name))
    table_rows, item_sections = _generate_item_sections(
        _sorted_dict_values(spec.items),
        app_name
    )

    headers = {
        'name': 'Name',
        'type': 'Type',
        'default': 'Default',
        'description': 'Description'
    }

    sections.append(
        build_markdown_table(
            headers,
            table_rows,
            ['name', 'type', 'default', 'description'],
        )
    )
    for item_section in item_sections:
        sections.append(item_section)

    return '\n'.join([section for section in sections])



# -*- coding: utf-8 -*-
"""Top-level package for Yapconf."""
import collections
import json
import re
import sys

import six

if sys.version_info.major < 3:
    from io import open
elif sys.version_info.major == 3:
    unicode = str

# Setup feature flags for use throughout the package.
yaml_support = True
etcd_support = True
kubernetes_support = True
redis_support = True

try:
    # We set safe_load, to be load because otherwise ruamel.yaml
    # will throw warnings. Since we don't want that to happen, and
    # we want our code to be the same whether or not PyYaml or
    # ruamel.yaml is installed.
    import ruamel.yaml as yaml
    yaml.load = yaml.safe_load
except ImportError:
    try:
        import yaml
    except ImportError:
        yaml = None
        yaml_support = False

try:
    from etcd import client as etcd_client
except ImportError as ex:
    etcd_client = None
    etcd_support = False


try:
    from kubernetes import client as kubernetes_client
except ImportError as ex:
    kubernetes_client = None
    kubernetes_support = False

from yapconf.exceptions import YapconfError  # noqa: E402
from yapconf.spec import YapconfSpec  # noqa: E402

__author__ = """Logan Asher Jones"""
__email__ = 'loganasherjones@gmail.com'
__version__ = '0.3.0'


FILE_TYPES = {'json', }
SUPPORTED_SOURCES = {
    'dict',
    'environment',
    'json',
}
ALL_SUPPORTED_SOURCES = {
    'dict',
    'environment',
    'etcd',
    'json',
    'kubernetes',
    'yaml',
}

if yaml_support:
    FILE_TYPES.add('yaml')
    SUPPORTED_SOURCES.add('yaml')

if etcd_support:
    SUPPORTED_SOURCES.add('etcd')

if kubernetes_support:
    SUPPORTED_SOURCES.add('kubernetes')

__all__ = ['YapconfSpec']


def change_case(s, separator='-'):
    """Changes the case to snake/kebab case depending on the separator.

    As regexes can be confusing, I'll just go through this line by line as an
    example with the following string: ' Foo2Boo_barBaz bat'

    1. Remove whitespaces from beginning/end. => 'Foo2Boo_barBaz bat'
    2. Replace all remaining spaces with underscores => 'Foo2Boo_barBaz_bat'
    3. Add underscores before capital letters => 'Foo2_Boo_bar_Baz_bat'
    4. Replace capital with lowercase => 'foo2_boo_bar_baz_bat'
    5. Replace underscores with the separator => 'foo2-boo-bar-baz-bat'

    Args:
        s (str): The original string.
        separator: The separator you want to use (default '-' for kebab case).

    Returns:
        A snake_case or kebab-case (depending on separator)
    """
    s = s.strip()
    no_spaces = re.sub(' ', '_', s)
    add_underscores = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', no_spaces)
    snake_case = re.sub('([a-z0-9])([A-Z])', r'\1_\2', add_underscores).lower()
    return re.sub('_', separator, snake_case)


def dump_data(data,
              filename=None,
              file_type='json',
              klazz=YapconfError,
              open_kwargs=None,
              dump_kwargs=None):
    """Dump data given to file or stdout in file_type.

    Args:
        data (dict): The dictionary to dump.
        filename (str, optional): Defaults to None. The filename to write
        the data to. If none is provided, it will be written to STDOUT.
        file_type (str, optional): Defaults to 'json'. Can be any of
        yapconf.FILE_TYPES
        klazz (optional): Defaults to YapconfError a special error to throw
        when something goes wrong.
        open_kwargs (dict, optional): Keyword arguments to open.
        dump_kwargs (dict, optional): Keyword arguments to dump.
    """

    _check_file_type(file_type, klazz)

    open_kwargs = open_kwargs or {'encoding': 'utf-8'}
    dump_kwargs = dump_kwargs or {}

    if filename:
        with open(filename, 'w', **open_kwargs) as conf_file:
            _dump(data, conf_file, file_type, **dump_kwargs)
    else:
        _dump(data, sys.stdout, file_type, **dump_kwargs)


def _dump(data, stream, file_type, **kwargs):
    if not kwargs and file_type == 'json':
        kwargs = {
            'sort_keys': True,
            'indent': 4,
            'ensure_ascii': False,
        }
    elif not kwargs and file_type == 'yaml':
        kwargs = {
            'default_flow_style': False,
            'encoding': 'utf-8'
        }

    if str(file_type).lower() == 'json':
        dumped = json.dumps(data, **kwargs)
        if isinstance(dumped, unicode):
            stream.write(dumped)
        else:
            stream.write(six.u(dumped))
    elif str(file_type).lower() == 'yaml':
        yaml.dump(data, stream, **kwargs)
    else:
        raise NotImplementedError('Someone forgot to implement dump for file '
                                  'type: %s' % file_type)


def load_file(filename,
              file_type='json',
              klazz=YapconfError,
              open_kwargs=None,
              load_kwargs=None):
    """Load a file with the given file type.

    Args:
        filename (str): The filename to load.
        file_type (str, optional): Defaults to 'json'. The file type for the
        given filename. Supported types are ``yapconf.FILE_TYPES```
        klazz (optional): The custom exception to raise if something goes
        wrong.
        open_kwargs (dict, optional): Keyword arguments for the open call.
        load_kwargs (dict, optional): Keyword arguments for the load call.

    Raises:
        klazz: If no klazz was passed in, this will be the ``YapconfError``

    Returns:
        dict: The dictionary from the file.
    """

    _check_file_type(file_type, klazz)

    open_kwargs = open_kwargs or {'encoding': 'utf-8'}
    load_kwargs = load_kwargs or {}

    data = None
    with open(filename, **open_kwargs) as conf_file:
        if str(file_type).lower() == 'json':
            data = json.load(conf_file, **load_kwargs)
        elif str(file_type).lower() == 'yaml':
            data = yaml.safe_load(conf_file.read())
        else:
            raise NotImplementedError('Someone forgot to implement how to '
                                      'load a %s file_type.' % file_type)

    if not isinstance(data, dict):
        raise klazz('Successfully loaded %s, but the result was '
                    'not a dictionary.' % filename)

    return data


def _check_file_type(file_type, klazz):
    if str(file_type).lower() == 'yaml' and yaml is None:
        raise klazz('You wanted to use a YAML file but the yaml module was '
                    'not loaded. Please install the yaml dependency via `pip '
                    'install yapconf[yaml]`.')

    if str(file_type).lower() not in FILE_TYPES:
        raise klazz('Invalid file type %s. Valid file types are %s' %
                    (file_type, FILE_TYPES))


def flatten(dictionary, separator='.', prefix=''):
    """Flatten the dictionary keys are separated by separator

    Arguments:
        dictionary {dict} -- The dictionary to be flattened.

    Keyword Arguments:
        separator {str} -- The separator to use (default is '.'). It will
        crush items with key conflicts.
        prefix {str} -- Used for recursive calls.

    Returns:
        dict -- The flattened dictionary.
    """
    new_dict = {}
    for key, value in dictionary.items():
        new_key = prefix + separator + key if prefix else key
        if isinstance(value, collections.MutableMapping):
            new_dict.update(flatten(value, separator, new_key))

        elif isinstance(value, list):
            new_value = []
            for item in value:
                if isinstance(item, collections.MutableMapping):
                    new_value.append(flatten(item, separator, new_key))
                else:
                    new_value.append(item)
            new_dict[new_key] = new_value

        else:
            new_dict[new_key] = value

    return new_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*- 
"""The setup script."""

from setuptools import find_packages, setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'six<2',
    'python-box<4',
    'watchdog',
]

extras = {
    'deploy': ['wheel', 'twine', ],
    'etcd': ['python-etcd', ],
    'develop': ['isort', 'watchdog', ],
    'docs': ['sphinx', 'sphinx_rtd_theme', ],
    'k8s': ['kubernetes', ],
    'test': [
        'codecov',
        'coverage',
        'flake8',
        'funcsigs',
        'kubernetes',
        'mock',
        'pluggy<0.7,>=0.5',
        'pytest',
        'pytest-lazy-fixture',
        'pytest-cov',
        'pytest-runner',
        'python-etcd',
        'ruamel.yaml',
        'tox',
    ],
    'yaml': ['ruamel.yaml'],
}

setup(
    name='yapconf',
    version='0.3.0',
    description="Yet Another Python Configuration",
    long_description=readme + '\n\n' + history,
    author="Logan Asher Jones",
    author_email='loganasherjones@gmail.com',
    url='https://github.com/loganasherjones/yapconf',
    packages=find_packages(include=['yapconf']),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='yapconf',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    test_suite='tests',
    tests_require=extras['test'],
    extras_require=extras,
)



# -*- coding: utf-8 -*-
import copy
import json
import os
import sys
import time
from argparse import ArgumentParser

import pytest
from mock import Mock, mock_open, patch
from ruamel import yaml

import yapconf
from yapconf.exceptions import (YapconfItemNotFound, YapconfLoadError,
                                YapconfSourceError, YapconfSpecError)
from yapconf.spec import YapconfSpec

if sys.version_info > (3,):
    long = int

original_env = None
original_yaml = yapconf.yaml
current_dir = os.path.abspath(os.path.dirname(__file__))


def setup_function(function):
    global original_env
    original_env = os.environ.copy()


def teardown_function(function):
    os.environ = original_env
    yapconf.yaml = original_yaml
    real_world_path = os.path.join(current_dir, 'files', 'real_world')
    tmp_path = os.path.join(real_world_path, 'tmp')

    json_file = os.path.join(real_world_path, 'config_to_change.json')
    yaml_file = os.path.join(real_world_path, 'config_to_change.yaml')
    original_data = {
        'database': {
            'host': '1.2.3.4',
            'name': 'myapp_prod',
            'port': 3306,
            'verbose': False,
        },
        'emoji': u'💩',
        'file': '/path/to/file.yaml',
        'ssl': {
            'private_key': 'blah',
            'public_key': 'blah',
        },
        'web_port': 443,
    }

    yapconf.dump_data(original_data, json_file, file_type='json')
    yapconf.dump_data(original_data, yaml_file, file_type='yaml')
    if os.path.exists(tmp_path):
        os.remove(tmp_path)


def test_simple_load_config(simple_spec):
    config = simple_spec.load_config({'my_string': 'some value',
                                      'my_int': 123,
                                      'my_long': long(123),
                                      'my_float': 123.45,
                                      'my_bool': False,
                                      'my_complex': 2j,
                                      })
    assert config['my_string'] == 'some value'
    assert config['my_int'] == 123
    assert config['my_long'] == long(123)
    assert config['my_float'] == 123.45
    assert not config['my_bool']
    assert config['my_complex'] == 2j


def test_load_list_config(spec_with_lists):
    config = spec_with_lists.load_config({
        'simple_list': ['a', 'b', 'c'],
        'top_list': [[1, 2, 3], [4, 5, 6]],
        'list_of_dictionaries': [
            {'foo': 'foo_value', 'bar': 'bar_value'},
            {'foo': 'foo_value2', 'bar': 'bar_value2'}
        ]
    })

    assert len(config['simple_list']) == 3
    assert 'a' in config['simple_list']
    assert 'b' in config['simple_list']
    assert 'c' in config['simple_list']

    assert len(config['top_list']) == 2
    assert len(config['top_list'][0]) == 3
    assert len(config['top_list'][1]) == 3

    assert len(config['list_of_dictionaries']) == 2
    assert 'foo' in config['list_of_dictionaries'][0]
    assert 'bar' in config['list_of_dictionaries'][0]
    assert 'foo' in config['list_of_dictionaries'][1]
    assert 'bar' in config['list_of_dictionaries'][1]


def test_nested_load_config(spec_with_dicts):
    config = spec_with_dicts.load_config(
        {
            'database': {
                'name': 'dbname',
                'host': 'dbhost',
                'port': 1234
            },
            'foo': {
                'bar': {'baz': 'baz_value'},
                'bat': True
            }
        }
    )
    assert config['database']['name'] == 'dbname'
    assert config['database']['host'] == 'dbhost'
    assert config['database']['port'] == 1234

    assert config['foo']['bar']['baz'] == 'baz_value'
    assert config['foo']['bat']


def test_spec_bad_file_type():
    with pytest.raises(YapconfSpecError):
        YapconfSpec({}, file_type='INVALID')


@pytest.mark.parametrize('bad_data', [
    (['THIS SHOULD BE A DICT']),
    ({'name': 'THIS SHOULD BE A DICT'}),
    ({'name': {'type': 'str', 'items': {'name2': {'type': 'str'}}}}),
    ({'name': {'type': 'list', 'error': 'it has not items entry'}}),
])
def test_invalid_specification(bad_data):
    with pytest.raises(YapconfSpecError):
        YapconfSpec(bad_data)


def test_load_config_fallbacks(simple_spec):
    config = simple_spec.load_config(
        {'my_string': 'some value'},
        {'my_int': 123},
        {'my_long': long(123)},
        {'my_float': 123.45},
        {'my_bool': False},
        {'my_complex': 2j}
    )
    assert config['my_string'] == 'some value'
    assert config['my_int'] == 123
    assert config['my_long'] == long(123)
    assert config['my_float'] == 123.45
    assert not config['my_bool']
    assert config['my_complex'] == 2j


@pytest.mark.parametrize('file_type,file_data,overrides,expected_value', [
    ('json', '{"foo": {"type": "str"}}', {"foo": "bar"}, {"foo": "bar"}),
    ('yaml',
     '''foo:
        type: str''',
     {'foo': 'bar'},
     {'foo': 'bar'})
])
def test_load_specification_from_file(file_type, file_data,
                                      overrides, expected_value):
    open_path = 'yapconf.open'
    with patch(open_path, mock_open(read_data=file_data)):
        spec = YapconfSpec('/path/to/specification', file_type=file_type)
        assert spec.load_config(overrides) == expected_value


def test_load_bad_specification_from_file():
    open_path = 'yapconf.open'
    with patch(open_path, mock_open(read_data="[]")):
        with pytest.raises(YapconfSpecError):
            YapconfSpec('/path/to/bad/spec', file_type='json')


def test_load_from_env(basic_spec):
    os.environ['FOO'] = 'foo_value'
    config = basic_spec.load_config('ENVIRONMENT')
    assert config['foo'] == 'foo_value'


@pytest.mark.parametrize('override', [
    ("Need another item",),
    ("label", "file", "invalid_file_type"),
    (["THIS SHOULD BE A DICT"])
])
def test_load_config_invalid_override(basic_spec, override):
    with pytest.raises(YapconfLoadError):
        basic_spec.load_config(override)


def test_load_config_yaml_not_supported(basic_spec):
    yapconf.yaml = None
    with pytest.raises(YapconfSourceError):
        basic_spec.load_config(('label', 'path/to/file', 'yaml'))


def test_load_config_nested_from_environment(spec_with_dicts):
    os.environ['FOO_BAR_BAZ'] = 'baz_value'
    config = spec_with_dicts.load_config(
        {
            'database': {'name': 'dbname', 'host': 'dbhost', 'port': 1234},
            'foo': {'bat': True}
        },
        'ENVIRONMENT'
    )
    value = spec_with_dicts.load_config(config)
    assert value['foo']['bar']['baz'] == 'baz_value'


def test_load_config_multi_part_dictionary(spec_with_dicts):
    config = spec_with_dicts.load_config(
        {'database': {'name': 'dbname'}},
        {'database': {'host': 'dbhost'}},
        {'database': {'port': 1234}},
        {'foo': {'bar': {'baz': 'baz_value'}, 'bat': True}}
    )
    value = spec_with_dicts.load_config(config)
    assert value['database'] == {'name': 'dbname',
                                 'host': 'dbhost',
                                 'port': 1234}


@patch('os.path.isfile', Mock(return_value=True))
def test_migrate_config_file():
    spec = YapconfSpec({'foo': {'bootstrap': True}})
    with patch('yapconf.open', mock_open(read_data='{}')):
        config = spec.migrate_config_file(
            '/path/to/file',
            create=False,
            include_bootstrap=False
        )
    assert config == {}


@patch('os.path.isfile', Mock(return_value=True))
def test_migrate_config_file_no_changes(basic_spec):
    open_path = 'yapconf.open'
    current_config = '{"foo": "bar"}'
    with patch(open_path, mock_open(read_data=current_config)):
        new_config = basic_spec.migrate_config_file('/path/to/file',
                                                    create=False)

    assert new_config == {"foo": "bar"}


@patch('os.path.isfile', Mock(return_value=False))
def test_migrate_config_file_does_not_exist_do_not_create(basic_spec):
    with pytest.raises(YapconfLoadError):
        basic_spec.migrate_config_file('/path/to/file', create=False)


@patch('os.path.isfile', Mock(return_value=False))
def test_migrate_config_file_does_not_exist_create(basic_spec):
    open_path = 'yapconf.open'
    with patch(open_path, mock_open()):
        new_config = basic_spec.migrate_config_file('/path/to/file',
                                                    create=True)

    assert new_config == {"foo": None}


@patch('os.path.isfile', Mock(return_value=False))
def test_migrate_config_file_create_yaml(basic_spec):
    open_path = 'yapconf.open'
    with patch('yapconf.yaml.dump') as dump_mock:
        with patch(open_path, mock_open()) as mock_file:
            new_config = basic_spec.migrate_config_file(
                '/path/to/file', create=True, output_file_type='yaml')
            dump_mock.assert_called_with(
                {"foo": None},
                mock_file(),
                default_flow_style=False,
                encoding='utf-8')

    assert new_config == {"foo": None}


@patch('os.path.isfile', Mock(return_value=True))
def test_migrate_config_file_always_update(basic_spec):
    open_path = 'yapconf.open'
    current_config = '{"foo": "bar"}'
    with patch(open_path, mock_open(read_data=current_config)):
        new_config = basic_spec.migrate_config_file('/path/to/file',
                                                    create=False,
                                                    always_update=True)

    assert new_config == {"foo": None}


@patch('os.path.isfile', Mock(return_value=True))
def test_migrate_config_file_update_previous_default():
    spec = YapconfSpec({'foo': {'default': 'baz',
                                'previous_defaults': ['bar']}})
    open_path = 'yapconf.open'
    current_config = '{"foo": "bar"}'
    with patch(open_path, mock_open(read_data=current_config)):
        new_config = spec.migrate_config_file('/path/to/file',
                                              create=False,
                                              update_defaults=True)
    assert new_config == {'foo': 'baz'}


def test_migrate_config_no_mock(tmpdir, basic_spec):
    new_path = tmpdir.join('config.json')
    with pytest.raises(YapconfLoadError):
        basic_spec.migrate_config_file(str(new_path), create=False)


def test_migrate_config_no_mock_create_file(tmpdir, basic_spec):
    new_path = tmpdir.join('config.json')
    basic_spec.migrate_config_file(str(new_path), create=True)
    assert json.load(new_path) == {"foo": None}


@pytest.mark.usefixtures('basic_spec')
@pytest.mark.parametrize('config', [
    ({'foo': None}),
    ({'foo': 'bar'}),
    ({'foo': u'bar'}),
    ({'foo': u'\U0001F4A9'}),
    ({'foo': u'💩'}),
    ({u'\U0001F4A9': 'foo'}),
    ({u'💩': 'foo'}),
    ({u'💩': u'💩'}),
])
def test_migrate_config_no_mock_existing_file(tmpdir, basic_spec, config):
    new_path = tmpdir.join('config.json')
    new_path.ensure()

    with new_path.open(mode='w') as fp:
        json.dump(config, fp)

    basic_spec.migrate_config_file(str(new_path), create=False)

    with new_path.open(mode='r') as fp:
        assert json.load(fp) == config


def test_get_item(basic_spec):
    assert basic_spec.get_item('foo') is not None
    assert basic_spec.get_item('no name match') is None


def test_update_defaults(basic_spec):
    assert basic_spec.defaults == {'foo': None}
    basic_spec.update_defaults({'foo': 'bar'})
    assert basic_spec.defaults == {'foo': 'bar'}


def test_update_defaults_bad_key(basic_spec):
    with pytest.raises(YapconfItemNotFound):
        basic_spec.update_defaults({"INVALID_KEY": "bar"})


def test_load_config_real_world(real_world_spec):
    parser = ArgumentParser(conflict_handler='resolve')
    real_world_spec.add_arguments(parser, bootstrap=True)
    cli_args = ['--file', '/path/to/file.yaml',
                '--web-port', '1234',
                '--ssl-public-key', '/path/to/public.crt',
                '--ssl-private-key', '/path/to/private.key',
                '--database-verbose'
                ]
    cli_values = vars(parser.parse_args(cli_args))
    bootstrap_config = real_world_spec.load_config(
        ('boostrap_cli', cli_values), 'ENVIRONMENT',
        bootstrap=True
    )
    os.environ['MY_APP_DATABASE_HOST'] = 'db_host_from_env'

    # Pretend we load this from bootstrap_config.file
    # Normally, you could just specify bootstrap_config.file
    # but because we are in a test, I'm not doing that.
    config_file_dict = {
        'ssl': {
            'public_key': '/path/to/public/from/file.crt',
            'private_key': '/path/to/private/from/file.key'
        },
        'database': {
            'name': 'name_from_config_file',
        },
        'web_port': 8080,
    }

    real_world_spec.add_arguments(parser, bootstrap=False)
    cli_values = vars(parser.parse_args(cli_args))

    config = real_world_spec.load_config(*[
        ('command line', cli_values),
        (bootstrap_config['file'], config_file_dict),
        'ENVIRONMENT'])

    assert config['file'] == '/path/to/file.yaml'
    assert config['web_port'] == 1234
    assert config['ssl']['public_key'] == '/path/to/public.crt'
    assert config['ssl']['private_key'] == '/path/to/private.key'
    assert config['database']['name'] == 'name_from_config_file'
    assert config['database']['host'] == 'db_host_from_env'
    assert config['database']['port'] == 3306
    assert config['database']['verbose']


@patch('os.path.isfile', Mock(return_value=True))
@pytest.mark.parametrize(
    'filename,expected,always_update,update_defaults,output_file_type', [
        (
            'default_current_config.yaml',
            pytest.lazy_fixture('default_current_config'),
            False,
            True,
            'yaml',
        ),
        (
            'default_previous_config.yaml',
            pytest.lazy_fixture('default_current_config'),
            False,
            True,
            'yaml',
        ),
        (
            'previous_config.yaml',
            {
                'emoji': u'🐍',
                'file': '/default/path/to/file.yaml',
                'web_port': 3000,
                'ssl': {
                    'private_key': '/previous/path/to/private.key',
                    'public_key': '/previous/path/to/public.crt'
                },
                'database': {
                    'name': 'myapp',
                    'host': 'localhost',
                    'port': 1234,
                    'verbose': False
                }
            },
            False,
            True,
            'json',
        ),
    ]
)
def test_real_world_migrations(
    real_world_spec,
    filename,
    expected,
    always_update,
    update_defaults,
    output_file_type
):
    real_world_path = os.path.join(current_dir, 'files', 'real_world')
    full_path = os.path.join(real_world_path, filename)
    tmp_config = os.path.join(real_world_path, 'tmp')

    new_config = real_world_spec.migrate_config_file(
        full_path,
        create=True,
        update_defaults=update_defaults,
        always_update=always_update,
        output_file_type=output_file_type,
        output_file_name=tmp_config
    )

    assert new_config == expected

    assert real_world_spec.load_config(tmp_config) == expected


@pytest.mark.parametrize('env_name,apply_prefix,env_prefix,key', [
    ('BG_HOST', True, 'BG_', 'BG_BG_HOST'),
    ('BG_HOST', False, 'BG_', 'BG_HOST'),
    ('HOST', True, 'BG_', 'BG_HOST'),
])
def test_env_names_with_prefixes(env_name, apply_prefix, env_prefix, key):
    spec = YapconfSpec(
        {
            'bg_host': {
                'type': 'str',
                'env_name': env_name,
                'apply_env_prefix': apply_prefix,
            }
        }, env_prefix=env_prefix)

    config = spec.load_config(('ENVIRONMENT', {key: 'host_value'}))
    assert config.bg_host == 'host_value'


@pytest.mark.usefixtures('simple_spec')
@pytest.mark.parametrize('key', [
    '/',
    '/foo/bar/',
])
def test_load_etcd(simple_spec, key):
    children = [
        Mock(key='%smy_string' % key, value='str_val'),
        Mock(key='%smy_int' % key, value='123'),
        Mock(key='%smy_long' % key, value='12341234'),
        Mock(key='%smy_float' % key, value='123.123'),
        Mock(key='%smy_bool' % key, value='true'),
        Mock(key='%smy_complex' % key, value='1j'),
    ]

    etcd_result = Mock(dir=True)
    etcd_result.children = children
    client = Mock(spec=yapconf.etcd_client.Client)
    client.read = Mock(return_value=etcd_result)
    simple_spec.add_source('etcd', 'etcd', client=client)
    config = simple_spec.load_config('etcd')
    assert config == {
        'my_string': 'str_val',
        'my_int': 123,
        'my_long': 12341234,
        'my_float': 123.123,
        'my_bool': True,
        'my_complex': 1j,
    }


@pytest.mark.parametrize('data,filename,file_type', [
    (None, 'unicode.json', 'json'),
    (None, 'unicode.yaml', 'yaml'),
    (pytest.lazy_fixture('example_data'), None, 'json'),
])
def test_load_from_source(
    example_spec,
    example_data,
    data,
    filename,
    file_type
):
    example_path = os.path.join(current_dir, 'files')
    if filename is not None:
        full_path = os.path.join(example_path, filename)
    else:
        full_path = None

    if data is not None:
        data = json.dumps(data)

    example_spec._file_type = file_type
    example_spec.add_source(
        'example',
        file_type,
        filename=full_path,
        data=data
    )

    config = example_spec.load_config('example')
    assert config == example_data


@pytest.mark.usefixtures('simple_spec')
@pytest.mark.parametrize('key,config_type,formatter', [
    (None, None, None),
    ('file.yaml', 'yaml', yaml.dump),
    ('file.json', 'json', json.dumps),
])
def test_load_kubernetes(simple_spec, key, config_type, formatter):
    data = {
        'my_string': 'str_val',
        'my_int': '123',
        'my_long': '12341234',
        'my_float': '123.123',
        'my_bool': 'True',
        'my_complex': '1j',
    }
    if formatter:
        data = formatter(data)

    if key:
        to_return = {key: data}
    else:
        to_return = data

    client = Mock(spec=yapconf.kubernetes_client.CoreV1Api)
    client.read_namespaced_config_map = Mock(return_value=Mock(data=to_return))
    simple_spec.add_source('kubernetes',
                           'kubernetes',
                           client=client,
                           config_type=config_type,
                           key=key)
    config = simple_spec.load_config('kubernetes')
    assert config == {
        'my_string': 'str_val',
        'my_int': 123,
        'my_long': 12341234,
        'my_float': 123.123,
        'my_bool': True,
        'my_complex': 1j,
    }


def test_defaults(fallback_spec):
    assert fallback_spec.defaults == {
        'defaults': {
            'str': 'default_str',
            'int': 123,
            'long': 123,
            'float': 123.123,
            'bool': True,
            'complex': 1j,
            'list': [1, 2, 3],
            'dict': {'foo': 'item_default'},
        },
        'str': None,
        'int': None,
        'long': None,
        'float': None,
        'bool': None,
        'complex': None,
        'list': None,
        'dict': {'foo': None},
    }


@pytest.mark.parametrize('configs,expected', [
    (
        [],
        {
            'defaults': {
                'str': 'default_str',
                'int': 123,
                'long': long(123),
                'float': 123.123,
                'bool': True,
                'complex': 1j,
                'list': [1, 2, 3],
                'dict': {'foo': 'item_default'},
            },
            'str': 'default_str',
            'int': 123,
            'long': long(123),
            'float': 123.123,
            'bool': True,
            'complex': 1j,
            'list': [1, 2, 3],
            'dict': {'foo': 'item_default'},
        }
    ),
    (
        [
            {
                'defaults': {
                    'str': 'fallback_str',
                    'int': 456,
                    'long': 456,
                    'float': 456.456,
                    'bool': True,
                    'complex': 4j,
                    'list': [4, 5, 6],
                    'dict': {'foo': 'fallback_item'}
                }
            },
            {
                'str': 'override_str',
                'int': 789,
                'long': 789,
                'float': 789.789,
                'bool': False,
                'complex': 7j,
                'list': [7, 8, 9],
                'dict': {'foo': 'override_item'}
            }
        ],
        {
            'defaults': {
                'str': 'fallback_str',
                'int': 456,
                'long': long(456),
                'float': 456.456,
                'bool': True,
                'complex': 4j,
                'list': [4, 5, 6],
                'dict': {'foo': 'fallback_item'},
            },
            'str': 'override_str',
            'int': 789,
            'long': long(789),
            'float': 789.789,
            'bool': False,
            'complex': 7j,
            'list': [7, 8, 9],
            'dict': {'foo': 'override_item'},
        }
    ),
    (
        [
            {
                'defaults': {
                    'str': 'fallback_str',
                    'int': 456,
                    'long': 456,
                    'float': 456.456,
                    'bool': True,
                    'complex': 4j,
                    'list': [4, 5, 6],
                    'dict': {'foo': 'fallback_item'}
                }
            },
            {
                'int': 789,
                'long': 789,
                'bool': False,
                'list': [7, 8, 9],
            }
        ],
        {
            'defaults': {
                'str': 'fallback_str',
                'int': 456,
                'long': long(456),
                'float': 456.456,
                'bool': True,
                'complex': 4j,
                'list': [4, 5, 6],
                'dict': {'foo': 'fallback_item'},
            },
            'str': 'fallback_str',
            'int': 789,
            'long': long(789),
            'float': 456.456,
            'bool': False,
            'complex': 4j,
            'list': [7, 8, 9],
            'dict': {'foo': 'fallback_item'},
        }
    ),
])
def test_fallbacks(fallback_spec, configs, expected):
    config = fallback_spec.load_config(*configs)
    assert config == expected


def test_generate_documentation_file(real_world_spec, tmpdir):
    new_path = tmpdir.join('real_world_doc.md')
    new_path.ensure()
    tmp_path = os.path.join(current_dir, 'files', 'real_world', 'doc.md')

    real_world_spec.add_source(
        'Source 1 Label', 'etcd', client=Mock(spec=yapconf.etcd_client.Client)
    )
    real_world_spec.add_source('Source 2 Label', 'dict', data={})
    real_world_spec.add_source('Source 3 Label', 'environment')
    real_world_spec.add_source('Source 4 Label', 'json', data={})
    real_world_spec.add_source(
        'Source 5 Label', 'json', filename='/path/to/file.json'
    )
    real_world_spec.add_source(
        'Source 6 Label',
        'kubernetes',
        client=Mock(spec=yapconf.kubernetes_client.CoreV1Api),
        key='key_name',
        name='config_map_name',
        namespace='config_map_namespace',
        config_type='json'
    )
    real_world_spec.add_source(
        'Source 7 Label', 'yaml', filename='/path/to/file.yaml'
    )
    real_world_spec.generate_documentation(
        'My App Name', output_file_name=str(new_path)
    )

    with new_path.open() as fp:
        generated_docs = fp.read()

    with open(tmp_path) as fp:
        expected = fp.read()

    assert generated_docs == expected


@pytest.mark.parametrize('spec,fq_name', [
    (pytest.lazy_fixture('real_world_spec'), 'file'),
    (pytest.lazy_fixture('real_world_spec'), 'emoji'),
    (pytest.lazy_fixture('real_world_spec'), 'ssl.private_key'),
    (pytest.lazy_fixture('spec_with_lists'), 'simple_list'),
])
def test_find_item(spec, fq_name):
    item = spec.find_item(fq_name)
    assert item.fq_name == fq_name


def test_spawn_watcher(simple_spec):
    simple_spec.add_source('label', 'dict', data={})
    mock_watch = Mock()
    simple_spec._sources['label'].watch = mock_watch
    simple_spec.spawn_watcher('label')
    assert mock_watch.call_count == 1


def test_spawn_watcher_error(simple_spec):
    with pytest.raises(YapconfSourceError):
        simple_spec.spawn_watcher('LABEL_NOT_DEFINED')


@pytest.mark.parametrize('label', [
    'label1',
    'label2',
    'label3',
])
def test_watchers(real_world_spec, label):
    original_data = {
        'database': {
            'host': '1.2.3.4',
            'name': 'myapp_prod',
            'port': 3307,
            'verbose': False,
        },
        'emoji': u'💩',
        'file': '/path/to/file.yaml',
        'ssl': {
            'private_key': 'blah',
            'public_key': 'blah',
        },
        'web_port': 443,
    }
    safe_data = copy.deepcopy(original_data)
    flags = {'overall': True, 'individual': True}
    real_world_path = os.path.join(current_dir, 'files', 'real_world')

    def overall_handler(old_config, new_config):
        flags['overall'] = False

    def indivual_handler(old_value, new_value):
        flags['individual'] = False

    def change_config(label):
        if label == 'label1':
            config = real_world_spec.load_config('label1')
            config.database.port += 1
            yapconf.dump_data(
                config.to_dict(),
                filename=yaml_filename,
                file_type='yaml'
            )

        elif label == 'label2':
            config = real_world_spec.load_config('label2')
            config.database.port += 1
            yapconf.dump_data(
                config.to_dict(),
                filename=json_filename,
                file_type='json'
            )

        elif label == 'label3':
            safe_data['database']['port'] += 1

    item = real_world_spec.find_item('database.port')
    item.watch_target = indivual_handler

    yaml_filename = os.path.join(real_world_path, 'config_to_change.yaml')
    json_filename = os.path.join(real_world_path, 'config_to_change.json')

    real_world_spec.add_source('label1', 'yaml', filename=yaml_filename)
    real_world_spec.add_source('label2', 'json', filename=json_filename)
    real_world_spec.add_source('label3', 'dict', data=safe_data)

    real_world_spec.spawn_watcher(label, target=overall_handler)

    change_config(label)

    wait_time = 0.0

    while any(flags.values()) and wait_time <= 30:
        time.sleep(0.25)
        wait_time += 0.25

    assert not all(flags.values())
